package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonObject
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.VertexConsumer
import com.mojang.blaze3d.vertex.VertexFormat
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.client.renderer.RenderStateShard
import net.minecraft.client.renderer.RenderType
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import org.lwjgl.opengl.GL11.GL_ALWAYS
import ru.dbotthepony.mc.otm.core.math.RGBAColor
import ru.dbotthepony.mc.otm.core.math.linearInterpolation
import java.util.concurrent.ConcurrentHashMap

sealed class AbstractMatterySprite {
	/**
	 * Expected image width in pixels, used in calculations
	 * and as default width argument in render methods
	 */
	abstract val width: Float

	/**
	 * Expected image height in pixels, used in calculations
	 * and as default height argument in render methods
	 */
	abstract val height: Float

	abstract val u0: Float
	abstract val v0: Float
	abstract val u1: Float
	abstract val v1: Float

	fun partialU(offset: Float): Float {
		return u0 + (offset / width) * (u1 - u0)
	}

	fun partialV(offset: Float): Float {
		return v0 + (offset / height) * (v1 - v0)
	}

	abstract val type: SpriteType

	open val winding: UVWindingOrder get() = UVWindingOrder.NORMAL

	abstract val texture: ResourceLocation

	/**
	 * See [ru.dbotthepony.mc.otm.client.render.clearDepth]
	 */
	fun clearDepth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
	) = ru.dbotthepony.mc.otm.client.render.clearDepth(stack, x, y, width, height)

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderWidth(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderHeight(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderHeight(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderHeight(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)


	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		val u1 = (u0 + linearInterpolation(width / this.width, 0f, u1 - u0)).coerceAtLeast(0f).coerceAtMost(1f)
		val v1 = (v0 + linearInterpolation(height / this.height, 0f, v1 - v0)).coerceAtLeast(0f).coerceAtMost(1f)

		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	protected fun uploadOnto(
		pose: PoseStack,
		builder: VertexConsumer,
		x: Float,
		y: Float,
		width: Float,
		height: Float,
		z: Float,
		color: RGBAColor?,
		u0: Float,
		v0: Float,
		u1: Float,
		v1: Float,
	) {
		val intColor = color?.toARGB() ?: 0
		val matrix = pose.last().pose()
		builder.vertex(matrix, x, y + height, z).also { if (color != null) it.color(intColor) }.uv(u0, v1).endVertex()
		builder.vertex(matrix, x + width, y + height, z).also { if (color != null) it.color(intColor) }.uv(u1, v1).endVertex()
		builder.vertex(matrix, x + width, y, z).also { if (color != null) it.color(intColor) }.uv(u1, v0).endVertex()
		builder.vertex(matrix, x, y, z).also { if (color != null) it.color(intColor) }.uv(u0, v0).endVertex()
	}

	@JvmOverloads
	fun uploadOnto(
		pose: PoseStack,
		builder: VertexConsumer,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		z: Float = 0f,
		color: RGBAColor? = null,
		winding: UVWindingOrder = this.winding,
	) {
		val (u0, v0, u1, v1) = winding.translate(u0, v0, u1, v1)
		uploadOnto(pose, builder, x, y, width, height, z, color, u0, v0, u1, v1)
	}

	@JvmOverloads
	fun uploadOntoPartial(
		pose: PoseStack,
		builder: VertexConsumer,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		z: Float = 0f,
		color: RGBAColor? = null,
		winding: UVWindingOrder = this.winding,
	) {
		val u1 = (u0 + linearInterpolation(width / this.width, 0f, u1 - u0)).coerceAtLeast(0f).coerceAtMost(1f)
		val v1 = (v0 + linearInterpolation(height / this.height, 0f, v1 - v0)).coerceAtLeast(0f).coerceAtMost(1f)
		val (u0_, v0_, u1_, v1_) = winding.translate(u0, v0, u1, v1)

		uploadOnto(pose, builder, x, y, width, height, z, color, u0_, v0_, u1_, v1_)
	}

	@JvmOverloads
	fun uploadOntoPartialColor(
		pose: PoseStack,
		builder: VertexConsumer,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		z: Float = 0f,
		color: RGBAColor = RGBAColor.WHITE,
		winding: UVWindingOrder = this.winding,
	) {
		uploadOntoPartial(pose, builder, x, y, width, height, z, color, winding)
	}

	@JvmOverloads
	fun uploadOntoColor(
		pose: PoseStack,
		builder: VertexConsumer,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		z: Float = 0f,
		color: RGBAColor = RGBAColor.WHITE,
		winding: UVWindingOrder = this.winding,
	) {
		uploadOnto(pose, builder, x, y, width, height, z, color, winding)
	}

	fun toJson(): JsonObject {
		return type.toActualJson(this)
	}

	fun toNetwork(buff: FriendlyByteBuf) {
		type.toActualNetwork(this, buff)
	}

	val renderTypeNoDepth by lazy {
		noDepthCache.computeIfAbsent(texture) {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(RenderStateShard.TextureStateShard(it, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setDepthTestState(RenderStateShard.DepthTestStateShard("always", GL_ALWAYS))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_no_depth",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				2048,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}
	}

	val renderTypeDepth by lazy {
		depthCache.computeIfAbsent(texture) {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(RenderStateShard.TextureStateShard(it, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_depth",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				2048,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}
	}

	val renderTypeWorld by lazy {
		worldCache.computeIfAbsent(texture) {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(RenderStateShard.TextureStateShard(it, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_world",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				8192,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}
	}

	companion object {
		private val noDepthCache = ConcurrentHashMap<ResourceLocation, RenderType>()
		private val depthCache = ConcurrentHashMap<ResourceLocation, RenderType>()
		private val worldCache = ConcurrentHashMap<ResourceLocation, RenderType>()
	}
}
