package ru.dbotthepony.mc.otm.core.collect

import java.util.Optional
import java.util.function.BinaryOperator
import java.util.function.Predicate
import java.util.stream.Collector

class FilteredIterator<T>(private val parent: Iterator<T>, private val predicate: Predicate<in T>) : Iterator<T> {
	private var foundValue: Any? = Companion

	override fun hasNext(): Boolean {
		if (foundValue === Companion) {
			while (parent.hasNext()) {
				val next = parent.next()

				if (predicate.test(next)) {
					foundValue = next
					return true
				}
			}

			return false
		}

		return true
	}

	@Suppress("unchecked_cast")
	override fun next(): T {
		var foundValue = foundValue

		if (foundValue === Companion) {
			if (!hasNext()) {
				throw NoSuchElementException()
			}

			foundValue = this.foundValue

			if (foundValue === Companion) {
				throw ConcurrentModificationException()
			}
		}

		this.foundValue = Companion
		return foundValue as T
	}

	private companion object
}

class MappingIterator<T, R>(private val parent: Iterator<T>, private val transform: (T) -> R) : Iterator<R> {
	override fun hasNext(): Boolean {
		return parent.hasNext()
	}

	override fun next(): R {
		return transform.invoke(parent.next())
	}
}

fun <T> Iterator<T>.filter(condition: Predicate<in T>): Iterator<T> = FilteredIterator(this, condition)
fun <T, R> Iterator<T>.map(mapper: (T) -> R): Iterator<R> = MappingIterator(this, mapper)
fun <T, R> Iterator<T>.map(mapper: java.util.function.Function<T, R>): Iterator<R> = MappingIterator(this, mapper::apply)

fun <T> Iterator<T>.reduce(identity: T, reducer: (T, T) -> T): T {
	var result = identity

	for (value in this) {
		result = reducer.invoke(result, value)
	}

	return result
}

fun <T> Iterator<T>.reduce(identity: T, reducer: BinaryOperator<T>): T = reduce(identity, reducer::apply)
fun <T> Iterator<T?>.filterNotNull(): Iterator<T> = filter { it != null } as Iterator<T>

fun <T> Iterator<T>.anyMatch(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (predicate.test(value)) {
			return true
		}
	}

	return false
}

fun <T> Iterator<T>.allMatch(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (!predicate.test(value)) {
			return false
		}
	}

	return true
}

fun <T> Iterator<T>.noneMatch(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (predicate.test(value)) {
			return false
		}
	}

	return true
}

fun <T, A, R> Iterator<T>.collect(collector: Collector<T, A, R>): R {
	val accumulator = collector.accumulator()
	val instance = collector.supplier().get()

	for (value in this) {
		accumulator.accept(instance, value)
	}

	return collector.finisher().apply(instance)
}

fun <T : Any> Iterator<T>.findFirst(): Optional<T> {
	if (hasNext()) {
		return Optional.of(next())
	}

	return Optional.empty<T>()
}

fun <T : Any> Iterator<T>.findAny() = findFirst()

fun <T> Iterator<T>.limit(limit: Long): Iterator<T> {
	require(limit > 0) { "Invalid limit $limit" }

	return object : Iterator<T> {
		var found = 0L

		override fun hasNext(): Boolean {
			return found < limit && this@limit.hasNext()
		}

		override fun next(): T {
			if (found >= limit) {
				throw NoSuchElementException()
			}

			return this@limit.next()
		}
	}
}

fun <T> Iterator<T>.skip(skip: Long): Iterator<T> {
	require(skip > 0) { "Invalid skip $skip" }

	return object : Iterator<T> {
		var found = skip

		override fun hasNext(): Boolean {
			while (this@skip.hasNext() && found > 0L) {
				found--
				this@skip.next()
			}

			return this@skip.hasNext()
		}

		override fun next(): T {
			if (!hasNext()) {
				throw NoSuchElementException()
			}

			return this@skip.next()
		}
	}
}

inline fun <T> Iterator<T>.forEach(action: (T) -> Unit) {
	for (value in this) {
		action.invoke(value)
	}
}

fun <T : Any> Iterator<T>.min(comparator: Comparator<T>): Optional<T> {
	if (!hasNext()) {
		return Optional.empty<T>()
	}

	var min = next()

	for (value in this) {
		if (comparator.compare(min, value) > 0) {
			min = value
		}
	}

	return Optional.of(min)
}

fun <T : Any> Iterator<T>.max(comparator: Comparator<T>): Optional<T> {
	if (!hasNext()) {
		return Optional.empty<T>()
	}

	var max = next()

	for (value in this) {
		if (comparator.compare(max, value) < 0) {
			max = value
		}
	}

	return Optional.of(max)
}
