package ru.dbotthepony.mc.otm.matter

import com.google.gson.JsonObject
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.set

class InsertAction : AbstractRegistryAction {
	val replaceIfExists: Boolean
	val comparePriority: Boolean

	constructor(
		key: ResourceLocation,
		matter: Decimal?,
		complexity: Double?,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		replaceIfExists: Boolean = false,
		comparePriority: Boolean = true,
	) : super(key = key, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		this.replaceIfExists = replaceIfExists
		this.comparePriority = comparePriority
	}

	constructor(
		tag: TagKey<Item>,
		matter: Decimal?,
		complexity: Double?,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		replaceIfExists: Boolean = false,
		comparePriority: Boolean = true,
	) : super(tag = tag, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		this.replaceIfExists = replaceIfExists
		this.comparePriority = comparePriority
	}

	constructor(json: JsonObject) : super(json) {
		this.replaceIfExists = json["replace_if_exists"]?.asBoolean ?: false
		this.comparePriority = json["compare_priority"]?.asBoolean ?: true
	}

	override fun toJson(): JsonObject {
		checkNotNull(matter) { "Missing matter value, which is required for insert action" }
		checkNotNull(complexity) { "Missing complexity value, which is required for insert action" }

		return super.toJson().also {
			it["action"] = "insert"
			it["replace_if_exists"] = replaceIfExists
			it["compare_priority"] = comparePriority
		}
	}

	override fun update(registry: MutableCollection<MutableEntry>, modifier: ResourceLocation) {
		checkNotNull(matter) { "Missing matter value, which is required for insert action (for $modifier)" }
		checkNotNull(complexity) { "Missing complexity value, which is required for insert action (for $modifier)" }

		check(!comparePriority || priority != null) { "If compare_priority is true then priority must not be null (for $modifier)" }

		if (!checkConditions()) {
			return
		}

		if (key != null) {
			if (replaceIfExists) {
				// search & replace in parallel, if possible
				val replaced = registry.stream().anyMatch {
					if (it is MutableKeyEntry && it.key == key) {
						if (!comparePriority || it.priority < priority!!) {
							it.matter = matter
							it.complexity = complexity
							it.priority = priority ?: 0
						}

						return@anyMatch true
					} else {
						return@anyMatch false
					}
				}

				if (!replaced) {
					registry.add(
						MutableKeyEntry(
							key,
							matter,
							complexity,
							priority ?: 0
						)
					)
				}
			} else {
				if (registry.stream().anyMatch { it is MutableKeyEntry && it.key == key }) {
					fail { "Value with key $key already exists" }
					return
				}

				registry.add(
					MutableKeyEntry(
						key,
						matter,
						complexity,
						priority ?: 0
					)
				)
			}
		} else {
			if (replaceIfExists) {
				// search & replace in parallel, if possible
				val replaced = registry.stream().anyMatch {
					if (it is MutableTagEntry && it.tag == tag) {
						if (!comparePriority || it.priority < priority!!) {
							it.matter = matter
							it.complexity = complexity
							it.priority = priority ?: 0
						}

						return@anyMatch true
					} else {
						return@anyMatch false
					}
				}

				if (!replaced) {
					registry.add(
						MutableTagEntry(
							tag!!,
							matter,
							complexity,
							priority ?: 0
						)
					)
				}
			} else {
				if (registry.stream().anyMatch { it is MutableTagEntry && it.tag == tag }) {
					fail { "Value with tag $tag already exists" }
					return
				}

				registry.add(
					MutableTagEntry(
						tag!!,
						matter,
						complexity,
						priority ?: 0
					)
				)
			}
		}
	}
}
